@@ -29,7 +29,8 @@ use libp2p_core::connection::{ConnectedPoint, ConnectionId, ListenerId};
 use libp2p_core::multiaddr::Multiaddr;
 use libp2p_core::PeerId;
 use libp2p_swarm::{
-    DialPeerCondition, NetworkBehaviour, NetworkBehaviourAction, NotifyHandler, PollParameters,
+    DialError, DialPeerCondition, IntoProtocolsHandler, NetworkBehaviour, NetworkBehaviourAction,
+    NotifyHandler, PollParameters,
 };
 use std::collections::{hash_map::Entry, HashMap, HashSet, VecDeque};
 use std::task::{Context, Poll};
@@ -45,7 +46,7 @@ pub struct Relay {
     /// [`Self::listeners`] or [`Self::listener_any_relay`].
     outbox_to_listeners: VecDeque<(PeerId, BehaviourToListenerMsg)>,
     /// Events that need to be yielded to the outside when polling.
-    outbox_to_swarm: VecDeque<NetworkBehaviourAction<RelayHandlerIn, ()>>,
+    outbox_to_swarm: VecDeque<NetworkBehaviourAction<(), RelayHandlerProto>>,
 
     /// List of peers the network is connected to.
     connected_peers: HashMap<PeerId, HashSet<ConnectionId>>,
@@ -201,6 +202,7 @@ impl NetworkBehaviour for Relay {
         peer: &PeerId,
         connection_id: &ConnectionId,
         _: &ConnectedPoint,
+        _: Option<&Vec<Multiaddr>>,
     ) {
         let is_first = self
             .connected_peers
@@ -295,42 +297,57 @@ impl NetworkBehaviour for Relay {
                     .push_back(NetworkBehaviourAction::NotifyHandler {
                         peer_id: *peer_id,
                         handler: NotifyHandler::Any,
-                        event: event,
+                        event,
                     });
             }
         }
     }
 
-    fn inject_dial_failure(&mut self, peer_id: &PeerId) {
-        if let Entry::Occupied(o) = self.listeners.entry(*peer_id) {
-            if matches!(o.get(), RelayListener::Connecting{ .. }) {
-                // By removing the entry, the channel to the listener is dropped and thus the
-                // listener is notified that dialing the relay failed.
-                o.remove_entry();
-            }
+    fn inject_dial_failure(
+        &mut self,
+        peer_id: Option<PeerId>,
+        _: Self::ProtocolsHandler,
+        error: &DialError,
+    ) {
+        if let DialError::DialPeerConditionFalse(
+            DialPeerCondition::Disconnected | DialPeerCondition::NotDialing,
+        ) = error
+        {
+            // Return early. The dial, that this dial was canceled for, might still succeed.
+            return;
         }
 
-        if let Some(reqs) = self.outgoing_relay_reqs.dialing.remove(peer_id) {
-            for req in reqs {
-                let _ = req.send_back.send(Err(OutgoingRelayReqError::DialingRelay));
+        if let Some(peer_id) = peer_id {
+            if let Entry::Occupied(o) = self.listeners.entry(peer_id) {
+                if matches!(o.get(), RelayListener::Connecting { .. }) {
+                    // By removing the entry, the channel to the listener is dropped and thus the
+                    // listener is notified that dialing the relay failed.
+                    o.remove_entry();
+                }
             }
-        }
 
-        if let Some(reqs) = self.incoming_relay_reqs.remove(peer_id) {
-            for req in reqs {
-                let IncomingRelayReq::DialingDst {
-                    src_peer_id,
-                    incoming_relay_req,
-                    ..
-                } = req;
-                self.outbox_to_swarm
-                    .push_back(NetworkBehaviourAction::NotifyHandler {
-                        peer_id: src_peer_id,
-                        handler: NotifyHandler::Any,
-                        event: RelayHandlerIn::DenyIncomingRelayReq(
-                            incoming_relay_req.deny(circuit_relay::Status::HopCantDialDst),
-                        ),
-                    })
+            if let Some(reqs) = self.outgoing_relay_reqs.dialing.remove(&peer_id) {
+                for req in reqs {
+                    let _ = req.send_back.send(Err(OutgoingRelayReqError::DialingRelay));
+                }
+            }
+
+            if let Some(reqs) = self.incoming_relay_reqs.remove(&peer_id) {
+                for req in reqs {
+                    let IncomingRelayReq::DialingDst {
+                        src_peer_id,
+                        incoming_relay_req,
+                        ..
+                    } = req;
+                    self.outbox_to_swarm
+                        .push_back(NetworkBehaviourAction::NotifyHandler {
+                            peer_id: src_peer_id,
+                            handler: NotifyHandler::Any,
+                            event: RelayHandlerIn::DenyIncomingRelayReq(
+                                incoming_relay_req.deny(circuit_relay::Status::HopCantDialDst),
+                            ),
+                        })
+                }
             }
         }
     }
@@ -340,6 +357,7 @@ impl NetworkBehaviour for Relay {
         peer: &PeerId,
         connection: &ConnectionId,
         _: &ConnectedPoint,
+        _: <Self::ProtocolsHandler as IntoProtocolsHandler>::Handler,
     ) {
         // Remove connection from the set of connections for the given peer. In case the set is
         // empty it will be removed in `inject_disconnected`.
@@ -396,15 +414,6 @@ impl NetworkBehaviour for Relay {
         }
     }
 
-    fn inject_addr_reach_failure(
-        &mut self,
-        _peer_id: Option<&PeerId>,
-        _addr: &Multiaddr,
-        _error: &dyn std::error::Error,
-    ) {
-        // Handled in `inject_dial_failure`.
-    }
-
     fn inject_listener_error(&mut self, _id: ListenerId, _err: &(dyn std::error::Error + 'static)) {
     }
 
@@ -472,10 +481,12 @@ impl NetworkBehaviour for Relay {
                                 src_connection_id: connection,
                             },
                         );
+                        let handler = self.new_handler();
                         self.outbox_to_swarm
                             .push_back(NetworkBehaviourAction::DialPeer {
                                 peer_id: dest_id,
                                 condition: DialPeerCondition::NotDialing,
+                                handler,
                             });
                     } else {
                         self.outbox_to_swarm
@@ -562,7 +573,7 @@ impl NetworkBehaviour for Relay {
         &mut self,
         cx: &mut Context<'_>,
         poll_parameters: &mut impl PollParameters,
-    ) -> Poll<NetworkBehaviourAction<RelayHandlerIn, Self::OutEvent>> {
+    ) -> Poll<NetworkBehaviourAction<Self::OutEvent, Self::ProtocolsHandler>> {
         if !self.outbox_to_listeners.is_empty() {
             let relay_peer_id = self.outbox_to_listeners[0].0;
 
@@ -623,7 +634,7 @@ impl NetworkBehaviour for Relay {
                     dst_peer_id,
                     send_back,
                 })) => {
-                    if let Some(_) = self.connected_peers.get(&relay_peer_id) {
+                    if self.connected_peers.get(&relay_peer_id).is_some() {
                         // In case we are already listening via the relay,
                         // prefer the primary connection.
                         let handler = self
@@ -668,6 +679,7 @@ impl NetworkBehaviour for Relay {
                         return Poll::Ready(NetworkBehaviourAction::DialPeer {
                             peer_id: relay_peer_id,
                             condition: DialPeerCondition::Disconnected,
+                            handler: self.new_handler(),
                         });
                     }
                 }
@@ -734,6 +746,7 @@ impl NetworkBehaviour for Relay {
                                 return Poll::Ready(NetworkBehaviourAction::DialPeer {
                                     peer_id: relay_peer_id,
                                     condition: DialPeerCondition::Disconnected,
+                                    handler: self.new_handler(),
                                 });
                             }
                         }
@@ -756,6 +769,7 @@ impl NetworkBehaviour for Relay {
 }
 
 #[derive(Debug)]
+#[allow(clippy::large_enum_variant)]
 pub enum BehaviourToListenerMsg {
     ConnectionToRelayEstablished,
     IncomingRelayedConnection {
